import types
from typing import Any, Dict, Optional, Sequence

import click
from typer import Typer
from typer.main import get_command as typer_get_command

from cliconf.arg_store import ARGS_STORE
from cliconf.command_name import get_command_name


def get_command(typer_instance: Typer) -> click.Command:
    """
    Extends typer's get_command function to modify the created click.Command instance
    to inspect the passed arguments and load from config.
    """
    command = typer_get_command(typer_instance)
    # Override the main function to load config
    command.main = types.MethodType(_cli_conf_main, command)
    return command


def _cli_conf_main(
    self: click.Command,
    args: Optional[Sequence[str]] = None,
    prog_name: Optional[str] = None,
    complete_var: Optional[str] = None,
    standalone_mode: bool = True,
    windows_expand_args: bool = True,
    **extra: Any,
) -> Any:
    """
    A modified version of click.Command's main function that records which arguments were passed
    """
    use_args = args or []
    func_name = prog_name or get_command_name(self.callback.__name__)  # type: ignore
    params = _create_passed_param_dict_from_command(self, func_name, use_args)
    # It seems typer always provides prog_name, but for safety calculate a fallback
    ARGS_STORE.add_command(func_name, use_args, params)
    return super(type(self), self).main(  # type: ignore
        args, func_name, complete_var, standalone_mode, windows_expand_args, **extra
    )


def _create_passed_param_dict_from_command(
    command: click.Command, prog_name: str, args: Sequence[str]
) -> Dict[str, Any]:
    context = command.make_context(prog_name, [*args])
    parser = command.make_parser(context)
    opts, _, param_order = parser.parse_args(args=[*args])
    # Reorder the opts dict to match the order of the command's params
    out_opts: Dict[str, Any] = {}
    for argument in param_order:
        if argument.name in opts:
            out_opts[argument.name] = opts[argument.name]
    return out_opts



from types import FunctionType
from typing import Any, Dict, Optional, Sequence, Type

from pyappconf import AppConfig, BaseConfig
from pydantic import create_model

from cliconf.arg_store import ARGS_STORE
from cliconf.command_name import get_command_name
from cliconf.ext_pyappconf import create_cli_base_config_class


def create_and_load_dynamic_config(
    func: FunctionType,
    func_args: Sequence[Any],
    func_kwargs: Dict[str, Any],
    settings: AppConfig,
    base_cls: Optional[Type[BaseConfig]] = None,
) -> BaseConfig:
    args_kwargs = dict(zip(func.__code__.co_varnames[1:], func_args))
    args_kwargs.update(func_kwargs)
    # Get user passed args from command line via args store
    args_store = ARGS_STORE[get_command_name(func.__name__)]
    user_kwargs = args_store.params
    base_cls = base_cls or create_cli_base_config_class(BaseConfig, settings)
    # Create a BaseConfig instance based off the function kwargs
    DynamicConfig = create_model(
        f"{func.__name__}_Config",
        __base__=base_cls,
        **args_kwargs,
        settings=settings,
        _settings=settings,
    )
    # Load the config, overriding with any user passed args
    return DynamicConfig.load(model_kwargs=user_kwargs)



from typing import Any, Dict, List, Sequence

from pydantic import BaseModel, Field


class CommandArgs(BaseModel):
    prog_name: str
    passed_args: List[Any]
    params: Dict[str, Any]


class ArgumentStore(BaseModel):
    """
    This class is used to store the arguments that are passed to the CLI.
    """

    commands: Dict[str, CommandArgs] = Field(default_factory=dict)

    def add_command(self, prog_name: str, args: Sequence[str], params: Dict[str, Any]):
        self.commands[prog_name] = CommandArgs(
            prog_name=prog_name, passed_args=list(args), params=params
        )

    def remove_command(self, prog_name: str):
        del self.commands[prog_name]

    def __getitem__(self, item):
        return self.commands[item]


ARGS_STORE = ArgumentStore()



import functools
from types import FunctionType
from typing import Callable, Optional, Type

from pyappconf import AppConfig, BaseConfig

from cliconf.dynamic_config import create_and_load_dynamic_config


def configure(
    settings: AppConfig, base_cls: Optional[Type[BaseConfig]] = None
) -> Callable:
    def actual_decorator(func: FunctionType):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Load the config, overriding with any user passed args
            config = create_and_load_dynamic_config(
                func, args, kwargs, settings, base_cls
            )
            return func(**config.dict(exclude={"settings"}))

        return wrapper

    return actual_decorator



import typer

from cliconf import configure
from cliconf.main import CLIConf
from tests.fixtures.app_settings import SETTINGS_ONE

single_command_cliconf = CLIConf(name="single_command")


@single_command_cliconf.command()
@configure(settings=SETTINGS_ONE)
def my_cli_func(
    a: str,
    b: int = typer.Argument(..., help="b help"),
    c: float = typer.Option(3.2, help="c help"),
):
    print(a, b, c)


if __name__ == "__main__":
    single_command_cliconf()



import shlex
from typing import Sequence

from click.testing import Result

from cliconf.main import CLIConf
from cliconf.testing import CLIRunner
from tests import ext_click
from tests.fixtures.cliconfs import single_command_cliconf

runner = CLIRunner()


class CLIRunnerException(Exception):
    pass


def run(instance: CLIConf, args: Sequence[str]) -> Result:
    result = runner.invoke(instance, args)
    if result.exit_code != 0:
        output = ext_click.result_to_message(result)
        command = shlex.join([instance.info.name, *args])
        raise CLIRunnerException(
            f"{command} with exited with code {result.exit_code}.\n{output}"
        )
    return result


def test_single_command_typer_reads_from_config():
    result = run(single_command_cliconf, ["a", "2"])
    assert result.stdout == "a 2 45.6\n"


def test_single_command_typer_reads_from_environment_over_config(monkeypatch):
    monkeypatch.setenv("MYAPP_C", "98.3")
    result = run(single_command_cliconf, ["a", "2"])
    assert result.stdout == "a 2 98.3\n"



from typing import Any

import typer

from cliconf.command import get_command


class CLIConf(typer.Typer):
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """
        Overrides Typer's __call__ method to use cliconf's get_command function.
        """
        return get_command(self)(*args, **kwargs)



from pyappconf import BaseConfig

from tests.fixtures.app_settings import SETTINGS_ONE


class ConfigOne(BaseConfig):
    a: str
    b: int
    c: float = 45.6

    _settings = SETTINGS_ONE


def generate_config_one():
    ConfigOne(a="a from config", b=1000).save()


if __name__ == "__main__":
    generate_config_one()



# This is the main settings file for package setup and PyPi deployment.
# Sphinx configuration is in the docsrc folder

# Main package name
PACKAGE_NAME = "cliconf"

# Directory name of package
PACKAGE_DIRECTORY = "cliconf"

# Name of Repo
REPO_NAME = "cliconf"

# Github username of the user which owns the repo
REPO_USERNAME = "nickderobertis"

# List of maintainers of package, by default the same user which owns the repo
# Pull requests raised by these maintainers without the "no auto merge" label will be automatically merged
REPO_MAINTAINERS = [
    REPO_USERNAME,
]

# Short description of the package
PACKAGE_SHORT_DESCRIPTION = "Framework that lets you write CLIs with Typer that can also be configured via py-app-conf"

# Long description of the package for PyPI
# Set to 'auto' to use README.md as the PyPI description
# Any other string will be used directly as the PyPI description
PACKAGE_DESCRIPTION = "auto"

# Author
PACKAGE_AUTHOR = "Nick DeRobertis"

# Author email
PACKAGE_AUTHOR_EMAIL = "derobertis.nick@gmail.com"

# Name of license for package
PACKAGE_LICENSE = "MIT"

# Classifications for the package, see common settings below
PACKAGE_CLASSIFIERS = [
    # How mature is this project? Common values are
    #   3 - Alpha
    #   4 - Beta
    #   5 - Production/Stable
    "Development Status :: 3 - Alpha",
    # Indicate who your project is intended for
    "Intended Audience :: Developers",
    # Specify the Python versions you support here. In particular, ensure
    # that you indicate whether you support Python 2, Python 3 or both.
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
]

# Add any third party packages you use in requirements here
PACKAGE_INSTALL_REQUIRES = [
    # Include the names of the packages and any required versions in as strings
    # e.g.
    # 'package',
    # 'otherpackage>=1,<2'
    "typer",
    "py-app-conf==1.3.0a6",
]

# Add any third party packages you use in requirements for optional features of your package here
# Keys should be name of the optional feature and values are lists of required packages
# E.g. {'feature1': ['pandas', 'numpy'], 'feature2': ['matplotlib']}
OPTIONAL_PACKAGE_INSTALL_REQUIRES = {}

# Packages added to Binder environment so that examples can be executed in Binder
# By default, takes this package (PACKAGE_NAME)
# everything the package requires (PACKAGE_INSTALL_REQUIRES) and everything
# that the package optionally requires (OPTIONAL_PACKAGE_INSTALL_REQUIRES) and adds them all to one list
# If a custom list is passed, it must include all the requirements for the Binder environment
BINDER_ENVIRONMENT_REQUIRES = list(
    set(
        PACKAGE_INSTALL_REQUIRES
        + [PACKAGE_NAME]
        + [
            package
            for package_list in OPTIONAL_PACKAGE_INSTALL_REQUIRES.values()
            for package in package_list
        ]
    )
)


# Sphinx executes all the import statements as it generates the documentation. To avoid having to install all
# the necessary packages, third-party packages can be passed to mock imports to just skip the import.
# By default, everything in PACKAGE_INSTALL_REQUIRES will be passed as mock imports, along with anything here.
# This variable is useful if a package includes multiple packages which need to be ignored.
DOCS_OTHER_MOCK_IMPORTS = [
    # Include the names of the packages as they would be imported, e.g.
    # 'package',
]

# Add any Python scripts which should be exposed to the command line in the format:
# CONSOLE_SCRIPTS = ['funniest-joke=funniest.command_line:main']
CONSOLE_SCRIPTS = ([],)

# Add any arbitrary scripts to be exposed to the command line in the format:
# SCRIPTS = ['bin/funniest-joke']
SCRIPTS = []

# Optional Google Analytics tracking ID for documentation
# Go to https://analytics.google.com/ and set it up for your documentation URL
# Set to None or empty string to not use this
GOOGLE_ANALYTICS_TRACKING_ID = ""

PACKAGE_URLS = {
    "Code": f"https://github.com/{REPO_USERNAME}/{REPO_NAME}",
    "Documentation": f"https://{REPO_USERNAME}.github.io/{REPO_NAME}",
}

# Url of logo
PACKAGE_LOGO_URL = ""

if __name__ == "__main__":
    # Store config as environment variables
    env_vars = dict(locals())
    # Imports after getting locals so that locals are only environment variables
    import shlex

    for name, value in env_vars.items():
        quoted_value = shlex.quote(str(value))
        print(f"export {name}={quoted_value};")



from pyappconf import AppConfig, ConfigFormats

from tests.config import CONFIGS_DIR

SETTINGS_ONE = AppConfig(
    app_name="MyApp",
    config_name="one",
    custom_config_folder=CONFIGS_DIR,
    default_format=ConfigFormats.YAML,
)



from typing import Optional, Type

from pyappconf import AppConfig, BaseConfig


def create_cli_base_config_class(
    base_cls: Type[BaseConfig], settings: Optional[AppConfig] = None
) -> Type[BaseConfig]:
    settings = settings or base_cls._settings
    prefix = _create_default_env_prefix(settings)

    class CLIBaseConfig(base_cls):  # type: ignore
        class Config:
            env_prefix = prefix

    return CLIBaseConfig


def _create_default_env_prefix(settings: AppConfig) -> str:
    return settings.app_name.replace("-", "_").replace(" ", "_").upper() + "_"
